---------------------------------------------------------------------------------------------
-- QHUnit Class
---------------------------------------------------------------------------------------------
QHUnit = {};
QHUnit.__index = QHUnit;

-- TODO: Add Element Container

function QHUnit:new(unitid)

	-- Metatable
	local self = {};
     	setmetatable(self, QHUnit);
     	
     	-- Unit Variables
	self.unitid = unitid;
	self.unittype = nil;
     	self.unitname = nil;
	self.unitrace = nil;
	self.unitclass = nil;
	self.unitlevel = nil;
	self.unithealth = nil;
	self.unithealthmax = nil;
	self.unitpower = nil;
	self.unitpowermax = nil;
	self.unitconnected = nil;
     	
     	-- Position Variables
	self.x = 0;
	self.y = 0;
	self.align = "TOPLEFT";
	self.relative = "TOPLEFT";
	self.anchor = UIParent;
	self.width = 0;
	self.height = 0;
	self.moving = nil;
	self.elements = {};
	self.numelements = 3;
	
	-- Frame Variables
	self.name = QH:Concat("QHUnit_", unitid);
	self.frame = CreateFrame("Frame", self.name, UIParent, "QHUNITTEMPLATE");
	
	-- Nameplate Variables
	self.nameplate = QHNameplate:new(QH:Concat(self.name, "_Name"), self.frame);
	self.nameplate.parent = self;
	
	-- Decurse Button Variables
	self.decurse = QHButton:new(QH:Concat(self.name, "_Decurse"), self.frame);
	self.decurse.parent = self;

	-- Buff Variables
	self.buffs = QHBuffs:new(QH:Concat(self.name, "_Buff"), self.frame);
	self.buffs.parent = self;

	-- Damage Bar Variables
	self.damagebar = QHBar:new(QH:Concat(self.name, "_DamageBar"), self.frame);
	self.damagebar.parent = self;
	
	-- Reference Variables
	self.obj = self;
	self.parent = nil;
	self.frame.obj = self;
	
	-- Call Update Function and Return
	self:Update();
	return self;
end

---------------------------------------------------------------------------------------------
-- QHUnit Methods
---------------------------------------------------------------------------------------------
function QHUnit:Clear()
	
	-- Unit Variables
	self.unittype = nil;
	self.unitname = nil;
	self.unitrace = nil;
	self.unitclass = nil;
	self.unitlevel = nil;
	self.unithealth = nil;
	self.unithealthmax = nil;
	self.unitpower = nil;
	self.unitpowermax = nil;
	
	self.nameplate:Clear();
	self.damagebar:Clear();
	self.decurse:Clear();
	self.buffs:Clear();
	
	-- Call Update
	self:Update();
end

function QHUnit:SetPosition(x, y, align, relative, anchor)
	self.x = x;
	self.y = y;
	self.align = align;
	self.relative = relative;
	self.anchor = anchor;
end

function QHUnit:ReSize()

	-- TODO: Clean Up

	-- Get Settings
	local name_pos = QH:GetSetting(QH:Concat(self.unittype, "_Name_Position"));
	local bar_pos = QH:GetSetting(QH:Concat(self.unittype, "_Bar_Position"));
	local dec_pos = QH:GetSetting(QH:Concat(self.unittype, "_Decurse_Position"));
	
	-- Calculate Total Width
	local w = 0;
	local count = 0;
	if name_pos and self.nameplate and self.nameplate.width then
		w = w + self.nameplate.width;
		count = count + 1;
	end
	if bar_pos and self.damagebar and self.damagebar.width then
		w = w + self.damagebar.width;
		count = count + 1;
	end
	if dec_pos and self.decurse and self.decurse.width then
		w = w + self.decurse.width;
		count = count + 1;
	end
	w = w + max((count - 1) * 5, 0);

	-- Calculate Largest Height
	local h = 0;
	if self.nameplate and self.nameplate.height and self.nameplate.height > h then
		h = self.nameplate.height;
	end
	if self.damagebar and self.damagebar.height and self.damagebar.height > h then
		h = self.damagebar.height;
	end
	if self.decurse and self.decurse.height and self.decurse.height > h then
		h = self.decurse.height;
	end
	
	self.width = w;
	self.height = h;
	return w, h;
end

function QHUnit:SortElements()
	
	-- Wipe Element Table
	table.wipe(self.elements);
	
	-- Get Element Positions
	local type = self:GetUnitType();
	local name_pos = QH:GetSetting(QH:Concat(type, "_Name_Position"));
	local bar_pos = QH:GetSetting(QH:Concat(type, "_Bar_Position"));
	local dec_pos = QH:GetSetting(QH:Concat(type, "_Decurse_Position"));
	
	-- TODO: Optional Order
	-- Insert Elements In Order
	if dec_pos then table.insert(self.elements, dec_pos, self.decurse); end
	if bar_pos then table.insert(self.elements, bar_pos, self.damagebar); end
	if name_pos then table.insert(self.elements, name_pos, self.nameplate); end
	
	-- Remove Gaps
	for n=1,self.numelements do
		if not self.elements[n] then
			self.elements[n] = "";
			table.remove(self.elements, n); 
		end
	end
end

function QHUnit:StartMoving(button)

	if self.obj.parent then
		self.obj.parent:StartMoving(button);
	elseif IsShiftKeyDown() then
		self.obj.moving = true;
		self.obj.frame:SetMovable(true);
		self.obj.frame:StartMoving();
	end
end

function QHUnit:StopMoving()

	if self.obj.parent then
		self.obj.parent:StopMoving();
	else
		self.obj.frame:StopMovingOrSizing();
		align, anchor, relative, x, y = self.frame:GetPoint();
		if not anchor then anchor = UIParent; end
		self.obj:SetPosition(x, y, align, relative, anchor);
		self.obj.moving = nil;
	end
end

function QHUnit:Hide()

	if self.frame:IsVisible() then self.frame:Hide(); end
end

function QHUnit:Show()
	
	if not self.frame:IsVisible() then self.frame:Show(); end
end

---------------------------------------------------------------------------------------------
-- QHUnit Unit Methods
---------------------------------------------------------------------------------------------
function QHUnit:LoadUnitInfo()

	if UnitExists(self.unitid) then

		local var = nil;
		
		var = UnitName(self.unitid);
		if var ~= self.unitname then self.unitchanged = true; end
		self.unitname = var;
		
		_, var = UnitRace(self.unitid);
		if var ~= self.unitrace then self.unitchanged = true; end
		self.unitrace = var;
		
		_, var = UnitClass(self.unitid);
		if var ~= self.unitclass then self.unitchanged = true; end
		self.unitclass = var;

		var = UnitLevel(self.unitid);
		if var ~= self.unitlevel then self.unitchanged = true; end
		self.unitlevel = var;
		
		var = UnitHealth(self.unitid);
		if var ~= self.unithealth then self.unitchanged = true; end
		self.unithealth = var;

		var = UnitHealthMax(self.unitid);
		if var ~= self.unithealthmax then self.unitchanged = true; end
		self.unithealthmax = var;
		
		var = UnitPower(self.unitid);
		if var ~= self.unitpower then self.unitchanged = true; end
		self.unitpower = var;

		var = UnitPowerMax(self.unitid);
		if var ~= self.unitpowerhmax then self.unitchanged = true; end
		self.unitpowerhmax = var;
		
		var = self:GetUnitType();
		if var ~= self.unittype then self.unitchanged = true; end
		self.unittype = var;
		
		var = UnitIsConnected(self.unitid);
		if var ~= self.unitconnected then self.unitchanged = true; end
		self.unitconnected = var;
	end
	return self.unitchanged;
end

function QHUnit:GetUnitType(unit)
	
	if not unit then unit = self.unitid; end
	
	-- Find Unit Type Of Name
	if unit == UnitName(unit) then
		
		-- Search Player/Target/Focus/Pet
		if UnitIsUnit(unit, "player") then
			return "Player";
		elseif UnitIsUnit(unit, "target") then
			return "Target";
		elseif UnitIsUnit(unit, "focus") then
			return "Focus";
		elseif UnitIsUnit(unit, "pet") then
			return "Pet";
		end
		
		-- Search Raid
		for i=1,40 do
			if UnitIsUnit(unit, QH:Concat("raid", i)) then
				return "Raid";
			end
		end
		
		-- Search Party
		for i=1,4 do
			if UnitIsUnit(unit, QH:Concat("party", i)) then
				return "Party";
			end
		end
	end
	
	-- Find Unit Type Of Unit
	if string.find(unit, "target", -6) then
		return "Target";
	elseif string.find(unit, "focus", -5) then
		return "Focus";
	elseif string.find(unit, "pet", -3) then
		return "Pet";
	elseif string.find(unit, "player") then
		return "Player";
	elseif string.find(unit, "party") then
		return "Party";
	elseif string.find(unit, "raid") then
		return "Raid";
	end
end

function QHUnit:GetUnitID(unit)
	
	if not unit then unit = self.unitid; end
	
	-- Search Player/Target/Focus/Pet
	if UnitIsUnit(unit, "player") then
		return "player";
	elseif UnitIsUnit(unit, "target") then
		return "target";
	elseif UnitIsUnit(unit, "focus") then
		return "focus";
	elseif UnitIsUnit(unit, "pet") then
		return "pet";
	end

	-- Search Raid
	for i=1,40 do
		if UnitIsUnit(unit, QH:Concat("raid", i)) then
			return "raid";
		end
	end

	-- Search Party
	for i=1,4 do
		if UnitIsUnit(unit, QH:Concat("party", i)) then
			return "party";
		end
	end
end

function QHUnit:GetUnitHeals(unit)

	local amount = 0;

	-- Estimate HoTs On Unit
	for i=1,40 do
	
		local name, rank, icon, count, type, duration, expires = UnitAura(unit,i,"HELPFUL|RAID");
	end
	
	-- Estimate Heals Being Cast On Unit
	
	return amount;
end

function QHUnit:UnitHasCureableDebuff(unit)

	if not unit then unit = self.unitid; end

	for i=1,40 do

		local name, rank, icon, count, type, duration, expires = UnitAura(unit,i,"HARMFUL|RAID");
		if(name) then return name, rank, icon, count, type, duration, expires; end
	end
end

function QHUnit:UnitInRange(unit)

	local spell = "";
	local _, playerclass = UnitClass("player");
	if(playerclass == "PALADIN") then
		spell = QH_LOCALE_PALADINRANGE;
	elseif(playerclass == "PRIEST") then
		spell = QH_LOCALE_PRIESTRANGE;
	elseif(playerclass == "DRUID") then
		spell = QH_LOCALE_DRUIDRANGE;
	elseif(playerclass == "SHAMAN") then
		spell = QH_LOCALE_SHAMANRANGE;
	end

	if(IsSpellInRange(spell,unit) == 1 or CheckInteractDistance(unit, 4)) then
		return true;
	end
end

---------------------------------------------------------------------------------------------
-- QHUnit Events
---------------------------------------------------------------------------------------------
function QHUnit:OnEvent(event, unit, ...)

	-- Convert Unit To String
	unit = tostring(unit);

	-- Player Events
	if event == "PLAYER_FOCUS_CHANGED" or event == "PLAYER_TARGET_CHANGED"
	or event == "PLAYER_ENTERING_WORLD" or not unit then 

		unit = "player";

	-- Party Events
	elseif event == "PARTY_MEMBER_ENABLE" or event == "PARTY_MEMBER_DISABLE" then

		unit = QHUnit:GetUnitID(unit);
	end

	-- Call Unit Update Method
	if unit then
		local unit_frame = getglobal(QH:Concat("QHUnit_", unit));
		--if unit_frame then unit_frame.obj:Update(); end
		if unit_frame then 
			unit_frame.obj.unitchanged = true; 
		end
	end
end

function QHUnit:Update(force_update)

	-- Load Unit Info
	self:LoadUnitInfo();
	
	-- Hide If Unit Doesn't Exist Or Anchor Is Hidden
	if not UnitExists(self.unitid) then --or (self.anchor and not self.anchor:IsVisible()) then
	
		-- Clean And Hide Nameplate
		if self.nameplate then
			self.nameplate:Clear();
			self.nameplate:Hide();
		end
		
		-- Clean And Hide Damagebar
		if self.damagebar then
			self.damagebar:Clear();
			self.damagebar:Hide();
		end
		
		-- Clean And Hide Decurse
		if self.decurse then
			self.decurse:Clear();
			self.decurse:Hide();
		end
		
		-- Clean And Hide Buffs
		if self.buffs then
			self.buffs:Clear();
			self.buffs:Hide();
		end
	else
		-- Update Position
		self:UpdatePosition();
		
		-- Update Range
		self:UpdateRange();

		-- Update Information
		if force_update or self.unitchanged then
		
			-- Update Nameplate
			self:UpdateNameplate();

			-- Update Damage Bar
			self:UpdateDamageBar();

			-- Update Decurse Button
			self:UpdateDecurse();
			
			-- Update Buffs
			self:UpdateBuffs();
		end
	end
	
	-- Mark Unit Updated
	self.unitchanged = false;
end

function QHUnit:UpdatePosition()

	-- Secure Code
	if not InCombatLockdown() then

		-- Sort Elements
		self:SortElements();

		-- Set Element Positions
		local x = 0;
		for index,value in pairs(self.elements) do
			if value then
				value:SetPosition(x, 0, "LEFT", "LEFT", self.frame);
				value:Show();
				x = x + value.width + 5;
			end
		end

		-- Verify Size
		self:ReSize();
		if not self.width then self.width = 0; end
		if not self.height then self.height = 0; end

		-- Verify Position
		if not self.x then self.x = 0; end
		if not self.y then self.y = 0; end
		if not self.align then self.align = "TOPLEFT"; end

		-- Frame Size
		self.frame:SetWidth(self.width);
		self.frame:SetHeight(self.height);

		-- Frame Position
		if not self.moving then
			self.frame:ClearAllPoints();
			self.frame:SetPoint(self.align, self.anchor, self.relative, self.x, self.y);
		end
	end
end

function QHUnit:UpdateRange()
	
	local range_check = QH:GetSetting("Bar_Range_Check");
	local name_range_alpha = QH:GetSetting("Name_Range_Alpha");
	local bar_range_alpha = QH:GetSetting("Bar_Range_Alpha");
	local decurse_range_alpha = QH:GetSetting("Decurse_Range_Alpha");
	
	if not range_check or self:UnitInRange(self.unitid) then
		name_range_alpha = 1.0;
		bar_range_alpha = 1.0;
		decurse_range_alpha = 1.0;
	end
	
	self.nameplate:UpdateAlpha(name_range_alpha);
	self.damagebar:UpdateAlpha(bar_range_alpha);
	self.decurse:UpdateAlpha(decurse_range_alpha);
end

function QHUnit:UpdateNameplate()

	if self.nameplate then
		
		-- Get Settings
		local name_pos = QH:GetSetting(QH:Concat(self.unittype, "_Name_Position"));
		local name_width = QH:GetSetting(QH:Concat(self.unittype, "_Name_Width"));
		local name_height = QH:GetSetting(QH:Concat(self.unittype, "_Name_Height"));
		local text_position = QH:GetSetting(QH:Concat(self.unittype, "_Name_Text_Position"));
		local background_color= QH:GetSetting("Name_Background_Color");

		if not name_pos then
		
			-- Clear and Hide
			self.nameplate:Clear();
			self.nameplate:Hide();
		else
			-- Set Size
			self.nameplate:SetSize(name_width, name_height);
			
			-- Set Label Position
			self.nameplate:SetTextPosition(text_position);
		
			-- Set Value
			if not self.unitname then
			
				-- Clear
				self.nameplate:Clear();
			else
				-- Label Value
				self.nameplate:SetText(self.unitname);

				-- Set Label Color
				if not UnitIsConnected(self.unitid) then
					self.nameplate:SetTextColor(0.5, 0.5, 0.5, 1.0);
				elseif self.unitclass then
					self.nameplate:SetTextColor(RAID_CLASS_COLORS[self.unitclass].r, RAID_CLASS_COLORS[self.unitclass].g, RAID_CLASS_COLORS[self.unitclass].b, 1.0);
				else
					self.nameplate:SetTextColor(1.0, 1.0, 1.0, 1.0);
				end

				-- Set Background
				self.nameplate:SetBackgroundColor(background_color.r, background_color.g, background_color.b, background_color.a);
			end
			
			-- Secure Code
			if not InCombatLockdown() then

				-- Show
				self.nameplate:Show();

				-- Set Button Attributes
				self.nameplate.button:Enable();
				self.nameplate.button:SetAttribute("unit", self.unitid);
				self.nameplate.button:SetAttribute("type1", "target");
				self.nameplate.button:SetAttribute("type2", "menu");
				--self.nameplate.button.menu = TargetFrameDropDown_Initialize;
				self.nameplate.button:RegisterForClicks('LeftButtonUp', 'RightButtonUp');
			end

			-- Update
			self.nameplate:Update();
		end
	end
end

function QHUnit:UpdateDamageBar()

	if self.damagebar then
	
		-- Get Bar Settings
		local bar_pos = QH:GetSetting(QH:Concat(self.unittype, "_Bar_Position"));
		local bar_width = QH:GetSetting(QH:Concat(self.unittype, "_Bar_Width"));
		local bar_height = QH:GetSetting(QH:Concat(self.unittype, "_Bar_Height"));
		local color_start = QH:GetSetting("Bar_Start_Color");
		local color_mid = QH:GetSetting("Bar_Mid_Color");
		local color_end = QH:GetSetting("Bar_End_Color");
		local color_hot = QH:GetSetting("Bar_Hot_Color");
		local label_show = QH:GetSetting("Bar_Label_Show");
		local color_label = QH:GetSetting("Bar_Label_Color");
		local background_show = QH:GetSetting("Bar_Background_Show");
		local background_color= QH:GetSetting("Bar_Background_Color");
		self.damagebar.sparkalways = QH:GetSetting("Bar_Always_Show_Spark");
		
		-- Get Button Settings
		local left_click = QH:GetSetting(QH:Concat(self.unittype, "_Left_Click"));
		local shift_left_click = QH:GetSetting(QH:Concat(self.unittype, "_Shift_Left_Click"));
		local ctrl_left_click = QH:GetSetting(QH:Concat(self.unittype, "_Ctrl_Left_Click"));
		local alt_left_click = QH:GetSetting(QH:Concat(self.unittype, "_Alt_Left_Click"));
		local right_click = QH:GetSetting(QH:Concat(self.unittype, "_Right_Click"));
		local shift_right_click = QH:GetSetting(QH:Concat(self.unittype, "_Shift_Right_Click"));
		local ctrl_right_click = QH:GetSetting(QH:Concat(self.unittype, "_Ctrl_Right_Click"));
		local alt_right_click = QH:GetSetting(QH:Concat(self.unittype, "_Alt_Right_Click"));
		local middle_click = QH:GetSetting(QH:Concat(self.unittype, "_Middle_Click"));
		local shift_middle_click = QH:GetSetting(QH:Concat(self.unittype, "_Shift_Middle_Click"));
		local ctrl_middle_click = QH:GetSetting(QH:Concat(self.unittype, "_Ctrl_Middle_Click"));
		local alt_middle_click = QH:GetSetting(QH:Concat(self.unittype, "_Alt_Middle_Click"));
		local button4_click = QH:GetSetting(QH:Concat(self.unittype, "_Button4_Click"));
		local shift_button4_click = QH:GetSetting(QH:Concat(self.unittype, "_Shift_Button4_Click"));
		local ctrl_button4_click = QH:GetSetting(QH:Concat(self.unittype, "_Ctrl_Button4_Click"));
		local alt_button4_click = QH:GetSetting(QH:Concat(self.unittype, "_Alt_Button4_Click"));
		local button5_click = QH:GetSetting(QH:Concat(self.unittype, "_Button5_Click"));
		local shift_button5_click = QH:GetSetting(QH:Concat(self.unittype, "_Shift_Button5_Click"));
		local ctrl_button5_click = QH:GetSetting(QH:Concat(self.unittype, "_Ctrl_Button5_Click"));
		local alt_button5_click = QH:GetSetting(QH:Concat(self.unittype, "_Alt_Button5_Click"));

		if not bar_pos then
		
			-- Clear And Hide
			self.damagebar:Clear();
			self.damagebar:Hide();
		else
			-- Set Size
			self.damagebar:SetSize(bar_width, bar_height);

			-- Set Value
			if not self.unithealth or not self.unithealthmax or not UnitIsConnected(self.unitid) then
			
				-- Clear
				self.damagebar:Clear();
			else

				-- Calculate Damage Percentage
				local damage = 1 - (self.unithealth / self.unithealthmax);
				
				-- Calculate Heal Estimates
				local heals = self:GetUnitHeals(self.unitid);

				-- Set Bar Positions
				self.damagebar:SetValueBack(self.unithealthmax - self.unithealth, self.unithealthmax);
				self.damagebar:SetValueFront(self.unithealthmax - self.unithealth - heals, self.unithealthmax);

				-- Set Bar Colors
				local red, green, blue;
				if damage < 0.5 then
					red = color_start.r + ((color_mid.r - color_start.r) * (damage / 0.5));
					green = color_start.g + ((color_mid.g - color_start.g) * (damage / 0.5));
					blue = color_start.b + ((color_mid.b - color_start.b) * (damage / 0.5));
				else
					red = color_mid.r + ((color_end.r - color_mid.r) * ((damage / 0.5) - 1));
					green = color_mid.g + ((color_end.g - color_mid.g) * ((damage / 0.5) - 1));
					blue = color_mid.b + ((color_end.b - color_mid.b) * ((damage / 0.5) - 1));
				end
				self.damagebar:SetColorFront(red, green, blue, range_alpha);
				self.damagebar:SetColorBack(color_hot.r, color_hot.g, color_hot.b, color_hot.a);
				self.damagebar:SetTextColorLeft(color_label.r, color_label.g, color_label.b, color_label.a);
				self.damagebar:SetTextColorRight(color_label.r, color_label.g, color_label.b, color_label.a);

				-- Set Left Label
				if label_show and UnitIsDeadOrGhost(self.unitid) then
					self.damagebar:SetTextLeft("Dead");
				elseif label_show and damage > 0 then
					self.damagebar:SetTextLeft(-(self.unithealthmax - self.unithealth));
				else
					self.damagebar:SetTextLeft("");
				end
				
				-- Set Right Label
				--self.damagebar:SetTextRight(self.unitid);
				
				-- Set Background
				if background_show then
					self.damagebar:SetBackgroundColor(background_color.r, background_color.g, background_color.b, background_color.a);
				else
					self.damagebar:SetBackgroundColor(0, 0, 0, 0);
				end
			end	
			
			-- Secure Code
			if not InCombatLockdown() then
			
				-- Show
				self.damagebar:Show();

				-- Set Button Attributes
				self:UpdateButton(self.damagebar.button, 1, left_click, shift_left_click, ctrl_left_click, alt_left_click, true);
				self:UpdateButton(self.damagebar.button, 2, right_click, shift_right_click, ctrl_right_click, alt_right_click, false);
				self:UpdateButton(self.damagebar.button, 3, middle_click, shift_middle_click, ctrl_middle_click, alt_middle_click, false);
				self:UpdateButton(self.damagebar.button, 4, button4_click, shift_button4_click, ctrl_button4_click, alt_button4_click, false);
				self:UpdateButton(self.damagebar.button, 5, button5_click, shift_button5_click, ctrl_button5_click, alt_button5_click, false);		
			end
			
			-- Update
			self.damagebar:Update();
		end
	end
end

function QHUnit:UpdateDecurse()

	if self.decurse then

		-- Get Settings
		local dec_show = QH:GetSetting("Decurse_Show");
		local dec_pos = QH:GetSetting(QH:Concat(self.unittype, "_Decurse_Position"));
		local dec_width = QH:GetSetting(QH:Concat(self.unittype, "_Decurse_Width"));
		local dec_height = QH:GetSetting(QH:Concat(self.unittype, "_Decurse_Height"));
		
		-- Get Button Settings
		local left_click = QH:GetSetting("Decurse_Left_Click");
		local right_click = QH:GetSetting("Decurse_Right_Click");
		local middle_click = QH:GetSetting("Decurse_Middle_Click");
		
		if not dec_pos or not dec_show then
		
			-- Clear And Hide
			self.decurse:Clear();
			self.decurse:Hide();
		else
		
			-- Set Size
			self.decurse:SetSize(dec_width, dec_height);

			-- Update Icon
			local debuff, rank, texture, count, debuff_type, duration, expires = self:UnitHasCureableDebuff();
			if not debuff then
			
				-- Clear
				self.decurse:Clear();
			else
				self.decurse:SetIcon(texture);
				if debuff_type then
					if dec_width > 20 then self.decurse:SetText(debuff_type);
					else self.decurse:SetText(string.sub(debuff_type, 1, 1)); end
				else
					self.decurse:SetText("");
				end
			end
			--self.decurse:SetIcon(ActionButton1Icon:GetTexture());
			
			-- Secure Code
			if not InCombatLockdown() then

				-- Show
				self.decurse:Show();

				-- Set Button Attributes
				self:UpdateButton(self.decurse.button, 1, left_click);
				self:UpdateButton(self.decurse.button, 2, right_click);
				self:UpdateButton(self.decurse.button, 3, middle_click);
			end
			
			-- Update
			self.decurse:Update();
		end
	end
end

function QHUnit:UpdateBuffs()

	if self.buffs then
	
		-- HACK: Adjust Frame Level
		if self.damagebar and self.damagebar.barfront then
			self.buffs.frame:SetFrameLevel(self.damagebar.barfront:GetFrameLevel());
			--if self.unitid == "player" then QH:Print(QH:Concat(self.damagebar.barfront:GetFrameLevel(), " : ", self.buffs.buffs[1].frame:GetFrameLevel())); end
		end
			
		-- Get Settings
		local buff_show = QH:GetSetting("Buff_Show");
		local buff_width = QH:GetSetting(QH:Concat(self.unittype, "_Buff_Width"));
		local buff_height = QH:GetSetting(QH:Concat(self.unittype, "_Buff_Height"));
		
		local buff_list = QH:GetSetting("Buff_List");

		if not buff_show then

			-- Clear And Hide
			self.buffs:Clear();
			self.buffs:Hide();
		else
			-- Set Size
			self.buffs:SetBuffSize(buff_width, buff_height);
			
			-- Set Position
			--self.buffs:SetPosition(-self.nameplate.label:GetStringWidth() - (2 * self.buffs.width), 0, "RIGHT", "RIGHT", self.frame);
			self.buffs:SetPosition(self.damagebar.labelleft:GetStringWidth(), 0, "LEFT", "LEFT", self.damagebar.frame);
			--self.damagebar.labelleft:GetStringWidth()
			
			-- Update Buffs
			self.buffs:Clear();
			local buff_index = 0;
			if buff_list then 
				for n = 1,self.buffs.buffcount do
					local buff_name = select(n, strsplit(";", buff_list));
					if buff_name then
						local buff, rank, texture, count, buff_type, duration, expires, caster = UnitAura(self.unitid, strtrim(buff_name));
						if buff and caster == "player" then 
							-- Update Icon
							buff_index = buff_index + 1;
							if count == 0 then count = ""; end
							self.buffs:SetBuff(buff_index, texture, count);
						end
					end
				end
			end			

			-- Secure Code
			if not InCombatLockdown() then

				-- Show
				self.buffs:Show();
			end

			-- Update
			self.buffs:Update();
		end
	end
end

function QHUnit:UpdateButton(button, index, action, shiftaction, ctrlaction, altaction, ondeath)

	--TODO: Generalize and make more efficient

	if button then
	
		local unit = self.unitid;

		button:Enable();
		button:SetAttribute("unit", unit);
		button:RegisterForClicks('LeftButtonUp', 'RightButtonUp', 'MiddleButtonUp', 'Button4Up', 'Button5Up');

		local _, playerclass = UnitClass("player");
		local resurrection = "";
		if(playerclass == "PRIEST") then
			resurrection = QH_LOCALE_PRIESTRES;
		elseif(playerclass == "DRUID") then
			resurrection = QH_LOCALE_DRUIDRES;
		elseif(playerclass == "PALADIN") then
			resurrection = QH_LOCALE_PALADINRES;
		elseif(playerclass == "SHAMAN") then
			resurrection = QH_LOCALE_SHAMANRES;
		end

		if(action == nil) then action = ""; end
		if(shiftaction == nil) then shiftaction = ""; end
		if(ctrlaction == nil) then ctrlaction = ""; end
		if(altaction == nil) then altaction = ""; end

		local name, texture, macrobody, localVar = GetMacroInfo(GetMacroIndexByName(string.gsub(action, "%!", "")));
		local shiftname, shifttexture, shiftmacrobody, shiftlocalVar = GetMacroInfo(GetMacroIndexByName(string.gsub(shiftaction, "%!", "")));
		local ctrlname, ctrltexture, ctrlmacrobody, ctrllocalVar = GetMacroInfo(GetMacroIndexByName(string.gsub(ctrlaction, "%!", "")));
		local altname, alttexture, altmacrobody, altlocalVar = GetMacroInfo(GetMacroIndexByName(string.gsub(altaction, "%!", "")));

		--Add Ressurect spell to Click upon death
		if(ondeath) then
		
			button:SetAttribute(QH:Concat("type",index), "macro");
			if(string.find(action, "!") and macrobody) then
				button:SetAttribute(QH:Concat("macrotext",index), QH:Concat("/cast [target=", unit, ", dead] ", resurrection, ";", string.char(13), string.gsub(macrobody, "unit", unit), ""));
			else
				button:SetAttribute(QH:Concat("macrotext",index), QH:Concat("/cast [target=", unit, ", dead] ", resurrection, "; [target=", unit, "] ", action, ";"));
			end

		--Normal Click
		else

			if(string.find(action, "!") and macrobody) then

				--Substitue "unit" variable in macro
				button:SetAttribute(QH:Concat("type",index), "macro");
				button:SetAttribute(QH:Concat("macrotext",index), string.gsub(macrobody, "unit", unit));
			else
				button:SetAttribute(QH:Concat("type",index), "spell");
				button:SetAttribute(QH:Concat("spell",index), string.gsub(action, "!", ""));
			end
		end

		--Shift-Click
		if(string.find(shiftaction, "!") and shiftmacrobody) then

			--Substitue "unit" variable in macro
			button:SetAttribute(QH:Concat("shift-type",index), "macro");
			button:SetAttribute(QH:Concat("shift-macrotext",index), string.gsub(shiftmacrobody, "unit", unit));

		else
			button:SetAttribute(QH:Concat("shift-type",index), "spell");
			button:SetAttribute(QH:Concat("shift-spell",index), string.gsub(shiftaction, "!", ""));
		end

		--Ctrl-Click
		if(string.find(ctrlaction, "!") and ctrlmacrobody) then

			--Substitue "unit" variable in macro
			button:SetAttribute(QH:Concat("ctrl-type",index), "macro");
			button:SetAttribute(QH:Concat("ctrl-macrotext",index), string.gsub(ctrlmacrobody, "unit", unit));

		else
			button:SetAttribute(QH:Concat("ctrl-type",index), "spell");
			button:SetAttribute(QH:Concat("ctrl-spell",index), string.gsub(ctrlaction, "!", ""));
		end

		--Alt-Click
		if(string.find(altaction, "!") and altmacrobody) then

			--Substitue "unit" variable in macro
			button:SetAttribute(QH:Concat("alt-type",index), "macro");
			button:SetAttribute(QH:Concat("alt-macrotext",index), string.gsub(altmacrobody, "unit", unit));

		else
			button:SetAttribute(QH:Concat("alt-type",index), "spell");
			button:SetAttribute(QH:Concat("alt-spell",index), string.gsub(altaction, "!", ""));
		end
	end
end